@@ -5,13 +5,6 @@ import warnings
 
 import attr
 
-# python traceback.TracebackException < 3.6.4 does not support unhashable exceptions
-# see https://github.com/python/cpython/pull/4014 for details
-if sys.version_info < (3, 6, 4):
-    exc_key = lambda exc: exc
-else:
-    exc_key = id
-
 ################################################################
 # MultiError
 ################################################################
@@ -114,6 +107,9 @@ def _filter_impl(handler, root_exc):
     preserved = set()
     new_root_exc = filter_tree(root_exc, preserved)
     push_tb_down(None, root_exc, preserved)
+    # Delete the local functions to avoid a reference cycle (see
+    # test_simple_cancel_scope_usage_doesnt_create_cyclic_garbage)
+    del filter_tree, push_tb_down
     return new_root_exc
 
 
@@ -132,6 +128,7 @@ class MultiErrorCatcher:
     def __exit__(self, etype, exc, tb):
         if exc is not None:
             filtered_exc = MultiError.filter(self._handler, exc)
+
             if filtered_exc is exc:
                 # Let the interpreter re-raise it
                 return False
@@ -149,6 +146,9 @@ class MultiErrorCatcher:
                 _, value, _ = sys.exc_info()
                 assert value is filtered_exc
                 value.__context__ = old_context
+                # delete references from locals to avoid creating cycles
+                # see test_MultiError_catch_doesnt_create_cyclic_garbage
+                del _, filtered_exc, value
 
 
 class MultiError(BaseException):
@@ -288,7 +288,6 @@ if have_tproxy:
 
         return tputil.make_proxy(controller, type(base_tb), base_tb)
 
-
 else:
     # ctypes it is
     import ctypes
@@ -339,7 +338,12 @@ else:
         c_new_tb.tb_lasti = base_tb.tb_lasti
         c_new_tb.tb_lineno = base_tb.tb_lineno
 
-        return new_tb
+        try:
+            return new_tb
+        finally:
+            # delete references from locals to avoid creating cycles
+            # see test_MultiError_catch_doesnt_create_cyclic_garbage
+            del new_tb, old_tb_frame
 
 
 def concat_tb(head, tail):
@@ -377,10 +381,12 @@ def traceback_exception_init(
     limit=None,
     lookup_lines=True,
     capture_locals=False,
+    compact=False,
     _seen=None,
+    **kwargs,
 ):
-    if _seen is None:
-        _seen = set()
+    if sys.version_info >= (3, 10):
+        kwargs["compact"] = compact
 
     # Capture the original exception and its cause and context as TracebackExceptions
     traceback_exception_original_init(
@@ -392,13 +398,19 @@ def traceback_exception_init(
         lookup_lines=lookup_lines,
         capture_locals=capture_locals,
         _seen=_seen,
+        **kwargs,
     )
 
+    seen_was_none = _seen is None
+
+    if _seen is None:
+        _seen = set()
+
     # Capture each of the exceptions in the MultiError along with each of their causes and contexts
     if isinstance(exc_value, MultiError):
         embedded = []
         for exc in exc_value.exceptions:
-            if exc_key(exc) not in _seen:
+            if id(exc) not in _seen:
                 embedded.append(
                     traceback.TracebackException.from_exception(
                         exc,
@@ -407,7 +419,8 @@ def traceback_exception_init(
                         capture_locals=capture_locals,
                         # copy the set of _seen exceptions so that duplicates
                         # shared between sub-exceptions are not omitted
-                        _seen=set(_seen),
+                        _seen=None if seen_was_none else set(_seen),
+                        **kwargs,
                     )
                 )
         self.embedded = embedded
@@ -415,7 +428,7 @@ def traceback_exception_init(
         self.embedded = []
 
 
-traceback.TracebackException.__init__ = traceback_exception_init
+traceback.TracebackException.__init__ = traceback_exception_init  # type: ignore
 traceback_exception_original_format = traceback.TracebackException.format
 
 
@@ -427,7 +440,7 @@ def traceback_exception_format(self, *, chain=True):
         yield from (textwrap.indent(line, " " * 2) for line in exc.format(chain=chain))
 
 
-traceback.TracebackException.format = traceback_exception_format
+traceback.TracebackException.format = traceback_exception_format  # type: ignore
 
 
 def trio_excepthook(etype, value, tb):
@@ -477,7 +490,7 @@ if sys.excepthook is sys.__excepthook__:
 # hook.
 #
 # More details: https://github.com/python-trio/trio/issues/1065
-if sys.excepthook.__name__ == "apport_excepthook":
+if getattr(sys.excepthook, "__name__", None) == "apport_excepthook":
     import apport_python_hook
 
     assert sys.excepthook is apport_python_hook.apport_excepthook
@@ -489,7 +502,7 @@ if sys.excepthook.__name__ == "apport_excepthook":
 
     fake_sys = TrioFakeSysModuleForApport()
     fake_sys.__dict__.update(sys.__dict__)
-    fake_sys.__excepthook__ = trio_excepthook
+    fake_sys.__excepthook__ = trio_excepthook  # type: ignore
     apport_python_hook.sys = fake_sys
 
     monkeypatched_or_warned = True
